-- Routines for reading and writing e-mail files (.EML or .MSG, see RFC 2822),
-- and Unix mailbox files (tested with Euphoria 2.4)
--
-- (c) Juergen Luethje <http://home.arcor.de/luethje/>
-- mbox 1.00 -- 2003, Nov 09
-- Freeware
-- License see "mbox_eng.txt"

include file.e
include wildcard.e


constant
   FALSE = 0, TRUE = 1,
   WSP = " \t",
   UNFOLD_CHARS = WSP & "\n\r",
   DELETE_IN_FILENAME = {"re","re[2]","re[3]","re[4]","aw","fw"},  -- like KorrNews
   FORBIDDEN_CHARS = ":\"/\\?*<>|",  -- characters not allowed in filenames (Windows)
   REPLACE_CHARS = ".'"


type boolean (object x)
   -- called by table_fetch(), parse_mbox()
   if integer(x) and (x = 0) or (x = 1) then
      return TRUE
   end if
   return FALSE
end type

type table (object x)
   -- called by table_append(), table_fetch(), parse_header(), header_string()
   integer n

   if sequence(x) then
      n = length(x)
      if (n >= 1) and (length(x[n]) = n-1) then
         return TRUE
      end if
   end if
   return FALSE
end type


function find_field (string header_s, integer start)
   -- return the beginning of the next header field
   -- called by parse_header()
   integer found

   found = finds('\n', header_s, start)
   while found do
      if found = length(header_s) or find(header_s[found+1], WSP) = 0 then
         return found
      end if
      found = finds('\n', header_s, found+1)
   end while
   return 0
end function


function find_field_char (integer char, string field_s)
   -- return the position of 'char', that delimits field name and body
   -- called by parse_header(), parse_mbox()
   sequence chars
   integer found

   -- find_any(char & WSP, field_s)
   chars = char & WSP
   found = 0
   for i = 1 to length(field_s) do
      if find(field_s[i], chars) then
         found = i
         exit
      end  if
   end for

   if found > 1 then
      found = verify(found, field_s, WSP)
      if found and (field_s[found] = char) then
         return found
      end if
   end if
   return 0
end function


function table_append (table theTable, object key, object x)
   -- append new entry
   -- called by parse_header()
   integer n

   n = length(theTable)
   theTable = append(theTable, append(theTable[n], key))
   theTable[n] = x
   return theTable
end function

function parse_header (string header_s)
   -- in : all header fields as string (with trailing \n)
   -- out: header as table
   -- called by mail_header()
   table header
   string field_s, name, body
   integer i, j, c

   header = {{}}     -- empty table
   i = 1
   while 1 do
      j = find_field(header_s, i)
      if j = 0 then exit end if
      field_s = header_s[i..j-1]
      c = find_field_char(':', field_s)
      name = trim_right(field_s[1..c-(c!=0)])
      body = trim_right(field_s[c+1..length(field_s)])
      header = table_append(header, name, body)
      i = j+1
   end while
   return header
end function

function mail_header (string mail_s)
   -- in : one mail as string
   -- out: header as table
   -- called by Subj_Date()
   string header_s
   integer f

   f = match("\n\n", mail_s)
   header_s = mail_s[1..f]
   return parse_header(header_s)
end function


function table_fetch (table theTable, object key, boolean match_case)
   -- return table item with indicated key,
   -- or default if key not found
   -- called by header_string()
   integer i

   if match_case = 0 then
      i = find(lower(key), lower(theTable[length(theTable)]))
   else
      i = find(key, theTable[length(theTable)])
   end if

   if i then
      return theTable[i]
   end if

   return ""           -- default
end function

function header_string (table header, string name)
   -- called by Subj_Date()
   return trim_left(table_fetch(header, name, FALSE))
end function

------------------------------------------------------------------------

constant HEX_CHARS = "0123456789ABCDEFabcdef"

function decode_QPencoded_text (string text)
   -- decode a Quoted-Printable 'encoded-text'
   -- [after RFC 2045, 2047]
   -- called by decode_header()
   sequence s, ret
   integer t, r, n, char

   t = 1
   r = 1
   n = length(text)
   ret = repeat(0, n)
   while t <= n do
      char = text[t]
      if char = '=' then
         t += 1
         if t <= n then
            if text[t] != '\n' then
               t += 1
               if t <= n then
                  if  find(text[t-1], HEX_CHARS)
                  and find(text[t],   HEX_CHARS) then
                     s = value('#' & upper(text[t-1..t]))
                     ret[r] = s[2]
                  else
                     ret[r] = '='
                     t -= 2
                  end if
                  r += 1
               end if
            end if
         end if
      elsif char = '_' then
         ret[r] = ' '
         r += 1
      else
         ret[r] = char
         r += 1
      end if
      t += 1
   end while
   return ret[1..r-1]
end function

function decode_header (string headerText)
   -- decode quoted-printable 'encoded-words' in a header field body
   -- [after RFC 2047]
   -- called by Subj_Date()
   sequence encoded
   string ret, ascii
   integer j, k

   j = match("=?", headerText)
   if j = 0 then return headerText end if

   ret = ""
   k = 0
   while 0 < j and j <= length(headerText)-2 do
      ascii = headerText[k+1..j-1]
      if verify(1, ascii, UNFOLD_CHARS) then
         ret &= ascii
      end if
      k = matchs("?=", headerText, j+2) + 1
      if (k > 3) and (headerText[k-3] = '?') and (lower(headerText[k-2]) = 'q') then
         k = matchs("?=", headerText, k+1) + 1
      end if
      if k = 1 then
         k = j - 1
         exit
      end if
      encoded = parse_any(headerText[j+2..k-2], "?")
      if (length(encoded) = 3) and equal(lower(encoded[2]), "q") then
         ret &= decode_QPencoded_text(encoded[3])
      else
         ret &= headerText[j..k]
      end if
      j = matchs("=?", headerText, k+1)
   end while
   return ret & headerText[k+1..length(headerText)]
end function

------------------------------------------------------------------------

global function From_line (string line)
   -- return TRUE, if 'line' begins with "From ", which is not a field name
   -- called by parse_mbox(), CanYouHandleThisFile()
   if match("From ", line) = 1 and line[verify(6,line,WSP)] != ':' then
      return TRUE
   end if
   return FALSE
end function

function trim_subject (string subject, sequence delInFilename)
   -- remove "Re:","Fw:" etc. *at the beginning* of the subject
   -- called by subject_to_filename()
   string temp
   integer f

   while 1 do
      subject = trim_left(subject)
      f = find(':', subject)
      if f = 0 then exit end if
      temp = trim_right(subject[1..f-1])
      if find(lower(temp), delInFilename) = 0 then exit end if
      subject = subject[f+1..length(subject)]
   end while
   return subject
end function

function subject_to_filename (string subject)
   -- called by Subj_Date()
   integer f

   if length(subject) then
      subject = trim_subject(subject, DELETE_IN_FILENAME)

      if length(subject) then
         -- replace forbidden characters
         for i = 1 to length(subject) do
            f = find(subject[i], FORBIDDEN_CHARS)
            if f > length(REPLACE_CHARS) then
               subject[i] = ' '
            elsif f != 0 then
               subject[i] = REPLACE_CHARS[f]
            end if
         end for
         subject = trim(subject)
         if length(subject) then
            return crunch(subject, UNFOLD_CHARS, ' ')
         else
            return "_"
         end if
      end if
   end if

   return "[no subject]"
end function

function Subj_Date (string mail_s)
   -- called by parse_mbox()
   sequence header
   string Subject, Date

   header = mail_header(mail_s)
   Subject = decode_header(header_string(header, "Subject"))
   Date = header_string(header, "Date")
   return {subject_to_filename(Subject), mail_date(Date)}
end function


------------------------------------------------------------------------
global function getln (integer fn)
   -- "wrapper" and replacement for gets()
   -- This function combines the advantages of reading lines
   -- o in text mode: A line looks always the same, regardless whether
   --   in the file it actually ends with "\n" or with "\r\n".
   -- o in binary mode: No problems with ASCII 26 (EOF).
   --
   -- in : number of a file, that was opened for reading in binary mode
   -- out: a line, as if it were read in text mode
   -- called by parse_mbox(), CanYouHandleThisFile()
   object line
   integer x

   line = gets(fn)
   if sequence(line) then
      x = length(line)-1
      if x > 0 and line[x] = '\r' then
         line[x] = '\n'
         return line[1..x]
      end if
   end if
   return line
end function
------------------------------------------------------------------------

global function parse_mbox (string file)
   -- return {subject, date, offset, size} for each mail in the given
   -- mbox file
   -- called by OpenArchive()
   object line, buffer
   sequence ret
   string mail_s, From_
   atom offset, fpos
   integer ifn
   boolean headerLinesOnly

   ifn = open(file, "rb")
   if ifn = -1 then return -1 end if

   line = getln(ifn)
   if atom(line) then          -- file size = 0 byte
      close(ifn)
      return {}
   end if

   if From_line(line) then
      offset = where(ifn)
      mail_s = ""
   else
      offset = 0
      mail_s = line
   end if
   ret = {}
   while 1 do
      line = getln(ifn)
      if atom(line) then exit end if

      if From_line(line) then
         -- Each message starts with a "From " line,
         -- that is immediately followed by the message header.
         fpos = where(ifn)
         From_ = line

         headerLinesOnly = FALSE
         buffer = getln(ifn)
         if atom(buffer) then exit end if
         if find_field_char(':', buffer) then
            while 1 do
               line = getln(ifn)
               if atom(line) then exit end if
               buffer &= line
               if find(line[1], WSP) = 0 and find_field_char(':', line) = 0 then
                  if equal(line, "\n") then
                     headerLinesOnly = TRUE
                  end if
                  exit
               end if
            end while
         end if

         if headerLinesOnly then
            ret = append(ret, Subj_Date(mail_s) & {offset, fpos-length(From_)-1 - offset})
            mail_s = buffer
            offset = fpos
         else
            mail_s &= From_ & buffer
         end if

      else
         mail_s &= line
      end if
   end while
   fpos = where(ifn)
   close(ifn)
   return append(ret, Subj_Date(mail_s) & {offset, fpos - offset})
end function

